# -*- coding: future_annotations -*-
import logging
from typing import TYPE_CHECKING

from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.singletons import nbs
from test.utils import assert_bool, make_safety_fixture, skipif_known_failing

if TYPE_CHECKING:
    from typing import Optional, Set

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, run_cell_ = make_safety_fixture(setup_cells=['from test.test_nested_symbols import DotDict'])
run_cell = run_cell_


def stale_detected():
    return nbs().test_and_clear_detected_flag()


def assert_detected(msg=''):
    assert_bool(stale_detected(), msg=msg)


def assert_not_detected(msg=''):
    assert_bool(not stale_detected(), msg=msg)


def lookup_symbols(val) -> Optional[Set[DataSymbol]]:
    safety = nbs()
    alias_set = {alias for alias in safety.aliases.get(id(val), []) if not alias.is_anonymous}
    if len(alias_set) == 0:
        return None
    return alias_set


def lookup_symbol(val) -> Optional[DataSymbol]:
    alias_set = lookup_symbols(val)
    if alias_set is None or len(alias_set) == 0:
        return None
    alias = None
    for alias in alias_set:
        # try to find one that isn't anonymous or
        # associated with anonymous namespace to avoid
        # e.g. <literal_sym_12345> in tests
        containing_ns = alias.containing_namespace
        if containing_ns is not None and containing_ns.is_anonymous:
            continue
        if alias.is_anonymous:
            continue
        return alias
    return alias


# ref: https://gist.github.com/golobor/397b5099d42da476a4e6
class DotDict(dict):
    """A dict with dot access and autocompletion.

    The idea and most of the code was taken from
    http://stackoverflow.com/a/23689767,
    http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
    http://stackoverflow.com/questions/2390827/how-to-properly-subclass-dict-and-override-get-set
    """
    def __init__(self, *a, **kw):
        dict.__init__(self)
        self.update(*a, **kw)
        self.__dict__ = self

    def __setattr__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError('This key is reserved for the dict methods.')
        dict.__setattr__(self, key, value)

    def __setitem__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError('This key is reserved for the dict methods.')
        dict.__setitem__(self, key, value)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self


def test_basic():
    run_cell('d = DotDict()')
    run_cell('d.x = DotDict()')
    run_cell('d.y = DotDict()')
    run_cell('d.x.a = 5')
    # run_cell('d.x.b = 6')
    run_cell('logging.info(d.x.a)')
    assert_not_detected()
    run_cell('logging.info(d.x)')
    assert_not_detected()
    run_cell('d.y.a = 7')
    # run_cell('d.y.b = 8')
    run_cell('logging.info(d.y.a)')
    assert_not_detected()
    run_cell('logging.info(d.y)')
    assert_not_detected()
    run_cell('x = d.x.a + 9')
    run_cell('y = d.y.a + 9')
    run_cell('d.y.a = 9')
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of changed `d.y.a`')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on changed `d.y.a`')
    run_cell('logging.info(d.x.a)')
    assert_not_detected()
    run_cell('logging.info(d.x)')
    assert_not_detected()
    run_cell('logging.info(d)')
    assert_not_detected()
    run_cell('d.y = 10')
    run_cell('logging.info(x)')
    assert_not_detected('`x` independent of changed `d.y`')
    run_cell('logging.info(y)')
    assert_detected('`y` depends on changed `d.y`')


def test_nested_readable_name():
    run_cell('d = DotDict()')
    run_cell('d.x = DotDict()')
    run_cell('d.x.a = 5')
    run_cell('d.x.b = 6')
    d_x_a = lookup_symbol(5)
    assert d_x_a.readable_name == 'd.x.a'
    d_x_b = lookup_symbol(6)
    assert d_x_b.readable_name == 'd.x.b'
    run_cell('d.y = DotDict()')
    run_cell('d.y.a = 7')
    run_cell('d.y.b = 8')
    d_y_a = lookup_symbol(7)
    assert d_y_a.readable_name == 'd.y.a'
    d_y_b = lookup_symbol(8)
    assert d_y_b.readable_name == 'd.y.b'


def test_nested_readable_name_dict_literal():
    run_cell('d = {"x": {"y": 42}}')
    d_x_y = lookup_symbol(42)
    assert d_x_y.readable_name == 'd[x][y]', 'got %s when expected d[x][y]' % d_x_y.readable_name


def test_nested_readable_name_list_literal():
    run_cell('lst = [0, [1, 2, 3]]')
    lst_1_1 = lookup_symbol(2)
    assert lst_1_1.readable_name == 'lst[1][1]', 'got %s when expected lst[1][1]' % lst_1_1.readable_name


def test_nested_readable_name_tuple_in_list():
    run_cell('lst = [0, (1, 2, 3)]')
    lst_1_1 = lookup_symbol(2)
    assert lst_1_1.readable_name == 'lst[1][1]', 'got %s when expected lst[1][1]' % lst_1_1.readable_name


def test_nested_readable_name_list_in_tuple():
    run_cell('lst = (0, [1, 2, 3])')
    lst_1_1 = lookup_symbol(2)
    assert lst_1_1.readable_name == 'lst[1][1]', 'got %s when expected lst[1][1]' % lst_1_1.readable_name


def test_nested_symbol_created_for_symbol_already_existing():
    run_cell('x = 42')
    run_cell('lst = [1, 2, [7, x, 8], 4]')
    x_syms = lookup_symbols(42)
    assert len(x_syms) == 2
    x_names = {sym.readable_name for sym in x_syms}
    assert 'x' in x_names, 'could not find `x` in set of names %s' % x_names
    assert 'lst[2][1]' in x_names, 'could not find `lst[2][1]` in set of names %s' % x_names


def test_list_append_extend():
    run_cell('lst = []')
    run_cell('lst.append(42)')
    run_cell('lst.extend([43, 44])')
    name = lookup_symbol(42).readable_name
    assert name == 'lst[0]', 'got %s' % name
    name = lookup_symbol(43).readable_name
    assert name == 'lst[1]', 'got %s' % name
    name = lookup_symbol(44).readable_name
    assert name == 'lst[2]', 'got %s' % name


def test_list_insert():
    run_cell('lst = [0, 1, 2, 4, 5, 6]')
    name = lookup_symbol(2).readable_name
    assert name == 'lst[2]', 'got %s' % name
    name = lookup_symbol(4).readable_name
    assert name == 'lst[3]', 'got %s' % name
    dsym = lookup_symbol(3)
    assert dsym is None
    run_cell('lst.insert(3, 3)')
    name = lookup_symbol(2).readable_name
    assert name == 'lst[2]', 'got %s' % name
    name = lookup_symbol(3).readable_name
    assert name == 'lst[3]', 'got %s' % name
    name = lookup_symbol(4).readable_name
    assert name == 'lst[4]', 'got %s' % name



# -*- coding: future_annotations -*-
from enum import Enum

from nbsafety.utils import fast


EMIT_EVENT = '_X5ix_NBSAFETY_EVT_EMIT'


class TraceEvent(Enum):
    init_cell = 'init_cell'

    before_stmt = 'before_stmt'
    after_stmt = 'after_stmt'

    attribute = 'attribute'
    subscript = 'subscript'

    before_complex_symbol = 'before_complex_symbol'
    after_complex_symbol = 'after_complex_symbol'

    before_lambda = 'before_lambda'
    after_lambda = 'after_lambda'

    before_call = 'before_call'
    after_call = 'after_call'
    argument = 'argument'
    before_return = 'before_return'
    after_return = 'after_return'

    before_dict_literal = 'before_dict_literal'
    after_dict_literal = 'after_dict_literal'
    before_list_literal = 'before_list_literal'
    after_list_literal = 'after_list_literal'
    before_set_literal = 'before_set_literal'
    after_set_literal = 'after_set_literal'
    before_tuple_literal = 'before_tuple_literal'
    after_tuple_literal = 'after_tuple_literal'

    dict_key = 'dict_key'
    dict_value = 'dict_value'
    list_elt = 'list_elt'
    set_elt = 'set_elt'
    tuple_elt = 'tuple_elt'

    before_assign_rhs = 'before_assign_rhs'
    after_assign_rhs = 'after_assign_rhs'

    line = 'line'
    call = 'call'
    return_ = 'return'
    exception = 'exception'

    # these are included for completeness but will probably not be used
    c_call = 'c_call'
    c_return = 'c_return'
    c_exception = 'c_exception'

    def __str__(self):
        return self.value

    def to_ast(self):
        return fast.Constant(self.value)



# -*- coding: future_annotations -*-
import ast
import copy
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Set
    from nbsafety.types import CellId


class StatementMapper(ast.NodeVisitor):
    def __init__(
        self,
        cell_id: Optional[CellId],
        line_to_stmt_map: Dict[int, ast.stmt],
        id_map: Dict[int, ast.AST],
        cell_id_by_ast_id: Dict[int, CellId],
        parent_map: Dict[int, ast.AST],
    ):
        self._cell_id: Optional[CellId] = cell_id
        self.line_to_stmt_map = line_to_stmt_map
        self.id_map = id_map
        self.cell_id_by_ast_id = cell_id_by_ast_id
        self.parent_map = parent_map
        self.traversal: List[ast.AST] = []

    def __call__(self, node: ast.AST) -> Dict[int, ast.AST]:
        # for some bizarre reason we need to visit once to clear empty nodes apparently
        self.visit(node)
        self.traversal.clear()
        
        self.visit(node)
        orig_traversal = self.traversal
        self.traversal = []
        self.visit(copy.deepcopy(node))
        copy_traversal = self.traversal
        orig_to_copy_mapping = {}
        for no, nc in zip(orig_traversal, copy_traversal):
            orig_to_copy_mapping[id(no)] = nc
            self.id_map[id(nc)] = nc
            self.cell_id_by_ast_id[id(nc)] = self._cell_id
            if isinstance(nc, ast.stmt):
                self.line_to_stmt_map[nc.lineno] = nc
                # workaround for python >= 3.8 wherein function calls seem
                # to yield trace frames that use the lineno of the first decorator
                for decorator in getattr(nc, 'decorator_list', []):
                    self.line_to_stmt_map[decorator.lineno] = nc
            nc_body = getattr(nc, 'body', [])
            try:
                for child in nc_body:
                    self.parent_map[id(child)] = nc
            except TypeError:
                self.parent_map[id(nc_body)] = nc
        return orig_to_copy_mapping

    def visit(self, node):
        self.traversal.append(node)
        for name, field in ast.iter_fields(node):
            if isinstance(field, ast.AST):
                self.visit(field)
            elif isinstance(field, list):
                for inner_node in field:
                    if isinstance(inner_node, ast.AST):
                        self.visit(inner_node)



# -*- coding: future_annotations -*-
import logging
from typing import TYPE_CHECKING

from nbsafety.singletons import nbs

if TYPE_CHECKING:
    from typing import Iterable, Set

    # avoid circular imports
    from nbsafety.data_model.data_symbol import DataSymbol

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class UpdateProtocol:
    def __init__(self, updated_sym: DataSymbol) -> None:
        self.updated_sym = updated_sym
        self.seen: Set[DataSymbol] = set()

    def __call__(self, new_deps: Set[DataSymbol], mutated: bool, propagate_to_namespace_descendents: bool) -> None:
        # in most cases, mutated implies that we should propagate to namespace descendents, since we
        # do not know how the mutation affects the namespace members. The exception is for specific
        # known events such as 'list.append()' or 'list.extend()' since we know these do not update
        # the namespace members.
        logger.warning(
            "updated sym %s (containing scope %s) with children %s",
            self.updated_sym,
            self.updated_sym.containing_scope,
            self.updated_sym.children_by_cell_position.values(),
        )
        directly_updated_symbols = nbs().aliases[self.updated_sym.obj_id] if mutated else {self.updated_sym}
        self._collect_updated_symbols_and_refresh_namespaces(
            directly_updated_symbols, propagate_to_namespace_descendents
        )
        logger.warning(
            'for symbol %s: mutated=%s; updated_symbols=%s', self.updated_sym, mutated, directly_updated_symbols
        )
        updated_symbols_with_ancestors = set(self.seen)
        logger.warning('all updated symbols for symbol %s: %s', self.updated_sym, updated_symbols_with_ancestors)
        nbs().updated_symbols |= self.seen
        for updated_sym in directly_updated_symbols:
            if not updated_sym.is_stale and updated_sym is not self.updated_sym:
                updated_sym.refresh()
        self.seen |= new_deps  # don't propagate to stuff on RHS
        for dsym in updated_symbols_with_ancestors:
            self._propagate_staleness_to_deps(dsym, skip_seen_check=True)

    def _collect_updated_symbols_and_refresh_namespaces(
        self, updated_symbols: Iterable[DataSymbol], refresh_descendent_namespaces: bool
    ) -> None:
        logger.warning('collecting updated symbols and namespaces for %s', updated_symbols)
        for dsym in updated_symbols:
            if dsym.is_import or dsym in self.seen:
                continue
            dsym.updated_timestamps.add(nbs().cell_counter())
            self.seen.add(dsym)
            containing_ns = dsym.containing_namespace
            if containing_ns is not None:
                logger.warning('containing scope for %s: %s; ids %s, %s', dsym, containing_ns, dsym.obj_id, containing_ns.obj_id)
                containing_ns.namespace_stale_symbols.discard(dsym)
                containing_ns.max_descendent_timestamp = nbs().cell_counter()
                self._collect_updated_symbols_and_refresh_namespaces(
                    nbs().aliases[containing_ns.obj_id], refresh_descendent_namespaces
                )
            if refresh_descendent_namespaces:
                dsym_ns = dsym.namespace
                if dsym_ns is not None:
                    self._collect_updated_symbols_and_refresh_namespaces(
                        dsym_ns.all_data_symbols_this_indentation(), refresh_descendent_namespaces
                    )

    def _propagate_staleness_to_namespace_parents(self, dsym: DataSymbol, skip_seen_check=False):
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        containing_ns = dsym.containing_namespace
        if containing_ns is None:
            return
        logger.warning("add %s to namespace stale symbols of %s", dsym, containing_ns)
        containing_ns.namespace_stale_symbols.add(dsym)
        for containing_alias in nbs().aliases[containing_ns.obj_id]:
            self._propagate_staleness_to_namespace_parents(containing_alias)

        for containing_alias in nbs().aliases[containing_ns.obj_id]:
            # do this in 2 separate loops to make sure all containing_alias are added to 'seen'
            # works around the issue when one alias depends on another
            for child in self._non_class_to_instance_children(containing_alias):
                logger.warning('propagate from namespace parent of %s to child %s', dsym, child)
                self._propagate_staleness_to_deps(child)

    def _non_class_to_instance_children(self, dsym):
        if self.updated_sym is dsym:
            for dep_introduced_pos, dsym_children in dsym.children_by_cell_position.items():
                if not nbs().settings.backwards_cell_staleness_propagation and dep_introduced_pos <= nbs().active_cell_position_idx:
                    continue
                yield from dsym_children
            return
        for dep_introduced_pos, dsym_children in dsym.children_by_cell_position.items():
            if not nbs().settings.backwards_cell_staleness_propagation and dep_introduced_pos <= nbs().active_cell_position_idx:
                continue
            for child in dsym_children:
                # Next, complicated check to avoid propagating along a class -> instance edge.
                # The only time this is OK is when we changed the class, which will not be the case here.
                child_namespace = child.namespace
                if child_namespace is not None and child_namespace.cloned_from is not None:
                    if child_namespace.cloned_from.obj_id == dsym.obj_id:
                        continue
                yield child

    def _propagate_staleness_to_namespace_children(self, dsym: DataSymbol, skip_seen_check=False):
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        self_ns = nbs().namespaces.get(dsym.obj_id, None)
        if self_ns is None:
            return
        for ns_child in self_ns.all_data_symbols_this_indentation(exclude_class=True):
            logger.warning('propagate from %s to namespace child %s', dsym, ns_child)
            self._propagate_staleness_to_deps(ns_child)

    def _propagate_staleness_to_deps(self, dsym: DataSymbol, skip_seen_check=False):
        if not skip_seen_check and dsym in self.seen:
            return
        self.seen.add(dsym)
        if dsym not in nbs().updated_symbols:
            if dsym.should_mark_stale(self.updated_sym):
                dsym.fresher_ancestors.add(self.updated_sym)
                dsym.required_timestamp = nbs().cell_counter()
                self._propagate_staleness_to_namespace_parents(dsym, skip_seen_check=True)
                self._propagate_staleness_to_namespace_children(dsym, skip_seen_check=True)
        for child in self._non_class_to_instance_children(dsym):
            logger.warning('propagate %s %s to %s', dsym, dsym.obj_id, child)
            self._propagate_staleness_to_deps(child)



# -*- coding: future_annotations -*-
import ast
import logging
import traceback
from typing import TYPE_CHECKING

from nbsafety.singletons import nbs
from nbsafety.tracing.ast_eavesdrop import AstEavesdropper
from nbsafety.tracing.stmt_inserter import StatementInserter
from nbsafety.tracing.stmt_mapper import StatementMapper

if TYPE_CHECKING:
    from typing import Optional
    from nbsafety.types import CellId


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class SafetyAstRewriter(ast.NodeTransformer):
    def __init__(self, cell_id: Optional[CellId]):
        self._cell_id: Optional[CellId] = cell_id

    def visit(self, node: ast.AST):
        try:
            mapper = StatementMapper(
                self._cell_id,
                nbs().statement_cache[nbs().cell_counter()],
                nbs().ast_node_by_id,
                nbs().cell_id_by_ast_id,
                nbs().parent_node_by_id,
            )
            orig_to_copy_mapping = mapper(node)
            # very important that the eavesdropper does not create new ast nodes for ast.stmt (but just
            # modifies existing ones), since StatementInserter relies on being able to map these
            node = AstEavesdropper(orig_to_copy_mapping).visit(node)
            node = StatementInserter(self._cell_id, orig_to_copy_mapping).visit(node)
        except Exception as e:
            nbs().set_ast_transformer_raised(e)
            traceback.print_exc()
            raise e
        return node



# -*- coding: future_annotations -*-
import logging

from nbsafety.singletons import nbs
from test.utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture(mark_typecheck_failures_unsafe=True)


def run_cell(cell, cell_id=None, **kwargs):
    """Mocks the `change active cell` portion of the comm protocol"""
    if cell_id is not None:
        nbs().handle({
            'type': 'change_active_cell',
            'active_cell_id': cell_id
        })
    run_cell_(cell, **kwargs)


def test_int_change_to_str_triggers_typecheck():
    run_cell('a = 1', 1)
    assert not nbs().get_cell_ids_needing_typecheck()
    run_cell('b = 2', 2)
    assert not nbs().get_cell_ids_needing_typecheck()
    run_cell('logging.info(a + b)', 3)
    assert not nbs().get_cell_ids_needing_typecheck()
    run_cell('b = "b"', 4)
    assert nbs().get_cell_ids_needing_typecheck() == {3}



# -*- coding: future_annotations -*-
import logging

from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.singletons import nbs
from test.utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, run_cell = make_safety_fixture()


def lookup_symbol(name: str) -> DataSymbol:
    ret = nbs().global_scope.lookup_data_symbol_by_name_this_indentation(name)
    assert ret is not None, 'got None for %s' % name
    return ret


def test_basic():
    run_cell('x = object()')
    assert lookup_symbol('x').get_ref_count() == 1
    run_cell('y = x')
    assert lookup_symbol('x').get_ref_count() == 2
    assert lookup_symbol('y').get_ref_count() == 2
    run_cell('del x')
    assert lookup_symbol('y').get_ref_count() == 1
    run_cell('y = None')
    # None has special semantics as it can mean that the symbol was gc'd
    # Right now (28/04/2021, hash 9099347) this isn't used anywhere but
    # that may change.
    assert lookup_symbol('y').get_ref_count() == -1



# -*- coding: future_annotations -*-
import ast
import logging
from typing import List, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import resolve_slice_to_constant
from nbsafety.analysis.mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.scope import NamespaceScope
from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    from typing import Dict, Optional, Set, Tuple, Union


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class ResolveRvalSymbols(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self):
        self.symbols: List[Optional[DataSymbol]] = []

    def __call__(self, node: ast.AST) -> Set[DataSymbol]:
        self.visit(node)
        return {sym for sym in self.symbols if sym is not None}

    def _push_symbols(self):
        return self.push_attributes(symbols=[])

    def visit_Name(self, node: ast.Name):
        self.symbols.extend(tracer().resolve_loaded_symbols(node))

    def visit_Tuple(self, node: ast.Tuple):
        self.visit_List_or_Tuple(node)

    def visit_List(self, node: ast.List):
        self.visit_List_or_Tuple(node)

    def visit_Dict(self, node: ast.Dict):
        resolved = tracer().resolve_loaded_symbols(node)
        if not resolved:
            # if id(node) not in tracer().node_id_to_loaded_literal_scope:
            # only descend if tracer failed to create literal symbol
            self.generic_visit(node.keys)
            self.generic_visit(node.values)
        else:
            self.symbols.extend(resolved)

    def visit_List_or_Tuple(self, node: Union[ast.List, ast.Tuple]):
        resolved = tracer().resolve_loaded_symbols(node)
        if not resolved:
            # if id(node) not in tracer().node_id_to_loaded_literal_scope:
            # only descend if tracer failed to create literal symbol
            self.generic_visit(node.elts)
        else:
            self.symbols.extend(resolved)

    def visit_AugAssign_or_AnnAssign(self, node):
        self.visit(node.value)

    def visit_AnnAssign(self, node):
        self.visit_AugAssign_or_AnnAssign(node)

    def visit_AugAssign(self, node):
        self.visit_AugAssign_or_AnnAssign(node)

    def visit_Call(self, node):
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            self.visit(node.func)
        self.symbols.extend(tracer().resolve_loaded_symbols(node.func))
        self.symbols.extend(tracer().resolve_loaded_symbols(node))
        self.generic_visit([node.args, node.keywords])

    def _get_attr_or_subscript_namespace(self, node: Union[ast.Attribute, ast.Subscript]) -> Optional[NamespaceScope]:
        with self._push_symbols():
            self.visit(node.value)
            symbols = self.symbols
        if len(symbols) != 1 or symbols[0] is None:
            return None
        return nbs().namespaces.get(symbols[0].obj_id, None)

    def visit_Attribute(self, node: ast.Attribute):
        if isinstance(node.value, ast.Call):
            self.visit(node.value)
        symbols = tracer().resolve_loaded_symbols(node)
        if len(symbols) > 0:
            self.symbols.extend(symbols)
            return
        # TODO: this path lacks coverage
        try:
            ns = self._get_attr_or_subscript_namespace(node)
            if ns is None:
                return
            dsym = ns.lookup_data_symbol_by_name_this_indentation(node.attr, is_subscript=False)
            if dsym is not None:
                self.symbols.append(dsym)
        except Exception as e:
            logger.warning("Exception occurred while resolving node %s: %s", ast.dump(node), e)

    def visit_Subscript(self, node: ast.Subscript):
        if isinstance(node.value, ast.Call):
            self.visit(node.value)
        symbols = tracer().resolve_loaded_symbols(node)
        with self._push_symbols():
            # add slice to RHS to avoid propagating to it
            self.visit(node.slice)
            symbols.extend(self.symbols)
        if len(symbols) > 0:
            self.symbols.extend(symbols)
            return
        # TODO: this path lacks coverage
        try:
            slice = resolve_slice_to_constant(node)
            if slice is None or isinstance(slice, ast.Name):
                return
            with self._push_symbols():
                self.visit(node.value)
                symbols = self.symbols
            if len(symbols) != 1 or symbols[0] is None:
                return
            ns = self._get_attr_or_subscript_namespace(node)
            if ns is None:
                return
            dsym = ns.lookup_data_symbol_by_name_this_indentation(slice, is_subscript=True)
            if dsym is None and isinstance(slice, int) and slice < 0:
                try:
                    dsym = ns.lookup_data_symbol_by_name_this_indentation(len(ns) + slice, is_subscript=True)
                except TypeError:
                    dsym = None
            if dsym is not None:
                self.symbols.append(dsym)
        except Exception as e:
            logger.warning("Exception occurred while resolving node %s: %s", ast.dump(node), e)

    def visit_keyword(self, node: ast.keyword):
        self.visit(node.value)

    def visit_Starred(self, node: ast.Starred):
        self.symbols.extend(tracer().resolve_loaded_symbols(node))

    def visit_Lambda(self, node):
        self.symbols.extend(tracer().resolve_loaded_symbols(node))

    def visit_GeneratorExp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_DictComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_ListComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_SetComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(self, node):
        to_append = set()
        for gen in node.generators:
            if isinstance(gen, ast.comprehension):
                with self._push_symbols():
                    self.visit(gen.iter)
                    self.visit(gen.ifs)
                    to_append |= set(self.symbols)
                with self._push_symbols():
                    self.visit(gen.target)
                    discard_set = set(self.symbols)
            else:
                with self._push_symbols():
                    self.visit(gen)
                    discard_set = set(self.symbols)
            to_append -= discard_set
        self.symbols.extend(to_append - discard_set)

    def visit_arg(self, node: ast.arg):
        self.symbols.extend(tracer().resolve_loaded_symbols(node.arg))

    def visit_For(self, node: ast.For):
        # skip body -- will have dummy since this visitor works line-by-line
        self.visit(node.iter)

    def visit_If(self, node: ast.If):
        # skip body here too
        self.visit(node.test)

    def visit_FunctionDef_or_AsyncFunctionDef(self, node: Union[ast.AsyncFunctionDef, ast.FunctionDef]):
        self.visit(node.args)
        self.generic_visit(node.decorator_list)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.visit_FunctionDef_or_AsyncFunctionDef(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        self.visit_FunctionDef_or_AsyncFunctionDef(node)

    def visit_ClassDef(self, node: ast.ClassDef):
        self.generic_visit(node.bases)
        self.generic_visit(node.decorator_list)

    def visit_With(self, node: ast.With):
        # skip body
        self.generic_visit(node.items)

    def visit_withitem(self, node: ast.withitem):
        self.visit(node.context_expr)

    def visit_ExceptHandler(self, node: ast.ExceptHandler):
        # important: this needs to skip the body
        if node.type is not None:
            self.symbols.extend(tracer().resolve_loaded_symbols(node.type))

    def visit_Import(self, node: ast.Import):
        pass

    def visit_ImportFrom(self, node: ast.ImportFrom):
        pass


def update_usage_info(symbols: Union[Optional[DataSymbol], Set[Optional[DataSymbol]]]):
    cell_counter = nbs().cell_counter()
    for sym in (symbols if symbols is not None and isinstance(symbols, set) else [symbols]):
        if sym is None:
            continue
        sym.last_used_cell_num = cell_counter
        logger.info('sym `%s` used in cell %d last updated in cell %d', sym, cell_counter, sym.timestamp)
        sym.timestamp_by_used_time[cell_counter] = sym.timestamp


def resolve_rval_symbols(node: Union[str, ast.AST], should_update_usage_info: bool = True) -> Set[DataSymbol]:
    if isinstance(node, str):
        node = ast.parse(node).body[0]
    rval_symbols = ResolveRvalSymbols()(node)
    if should_update_usage_info:
        update_usage_info(rval_symbols)
    return rval_symbols



# -*- coding: future_annotations -*-
import ast
import sys
# from .utils import skipif_known_failing

from nbsafety.analysis.live_refs import compute_live_dead_symbol_refs


def _remove_callpoints(symbols):
    return set(sym for sym in symbols if isinstance(sym, str))


def test_simple():
    live, dead = compute_live_dead_symbol_refs("""
x = 5
print(foo, x)""")
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'foo', 'print'}
    assert dead == {'x'}


def test_function_body():
    fbody = ast.parse("""
def func():
    y = 42
    print(foo, bar, baz, x)
    x = 5
""").body[0].body
    live, dead = compute_live_dead_symbol_refs(fbody)
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'foo', 'bar', 'baz', 'x', 'print'}
    assert dead == {'x', 'y'}


def test_comprehension_with_killed_elt():
    live, dead = compute_live_dead_symbol_refs('[x for y in range(10) for x in range(11)]')
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'range'}
    assert dead == {'x', 'y'}


def test_comprehension_with_live_elt():
    live, dead = compute_live_dead_symbol_refs('[x for y in range(10) for _ in range(11)]')
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'x', 'range'}
    assert dead == {'y', '_'}


def test_subscript_is_live():
    live, dead = compute_live_dead_symbol_refs('foo[bar] = baz')
    live, dead = _remove_callpoints(live), _remove_callpoints(dead)
    assert live == {'bar', 'baz'}


if sys.version_info >= (3, 8):
    def test_walrus():
        live, dead = compute_live_dead_symbol_refs("""
if (y := (x := x + 1) + 1) > 0:
    z = y + 1
""")
        assert live == {'x'}
        assert dead == {'y', 'z'}



# -*- coding: future_annotations -*-
import ast
import builtins
from collections import defaultdict
from contextlib import contextmanager
import functools
import logging
import sys
from typing import cast, TYPE_CHECKING

import astunparse
from IPython import get_ipython

from nbsafety import singletons
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.scope import Scope, NamespaceScope
from nbsafety.run_mode import SafetyRunMode
from nbsafety.singletons import nbs
from nbsafety.tracing.mutation_event import ArgMutate, ListAppend, ListExtend, ListInsert, StandardMutation
from nbsafety.tracing.symbol_resolver import resolve_rval_symbols
from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
from nbsafety.tracing.trace_stack import TraceStack
from nbsafety.tracing.trace_stmt import TraceStatement
from nbsafety.tracing.utils import match_container_obj_or_namespace_with_literal_nodes

if TYPE_CHECKING:
    from typing import Any, Callable, DefaultDict, Dict, List, Optional, Set, Tuple, Type, Union
    from types import FrameType
    from nbsafety.tracing.mutation_event import MutationEvent
    from nbsafety.types import SupportedIndexType
    AttrSubVal = SupportedIndexType
    NodeId = int
    ObjId = int
    MutationCandidate = Tuple[Tuple[int, ObjId, Optional[str]], MutationEvent, Set[DataSymbol], List[Any]]
    Mutation = Tuple[int, MutationEvent, Set[DataSymbol], List[Any]]
    SavedStoreData = Tuple[NamespaceScope, Any, AttrSubVal, bool]
    SavedDelData = Tuple[NamespaceScope, AttrSubVal, bool]
    SavedComplexSymbolLoadData = Tuple[Tuple[NamespaceScope, Any], Tuple[AttrSubVal, bool]]


logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


ARG_MUTATION_EXCEPTED_MODULES = {
    'alt',
    'altair',
    'display',
    'logging',
    'matplotlib',
    'pyplot',
    'plot',
    'plt',
    'seaborn',
    'sns',
    'widget',
}


def _finish_tracing_reset():
    # do nothing; we just want to trigger the newly reenabled tracer with a 'call' event
    pass


class BaseTraceManager(singletons.TraceManager):

    _MANAGER_CLASS_REGISTERED = False
    EVENT_HANDLERS_PENDING_REGISTRATION: DefaultDict[TraceEvent, List[Callable[..., Any]]] = defaultdict(list)
    EVENT_HANDLERS_BY_CLASS: Dict[Type[BaseTraceManager], DefaultDict[TraceEvent, List[Callable[..., Any]]]] = {}

    EVENT_LOGGER = logging.getLogger('events')
    EVENT_LOGGER.setLevel(logging.WARNING)

    def __init__(self):
        if not self._MANAGER_CLASS_REGISTERED:
            raise ValueError(
                f'class not registered; use the `{register_trace_manager_class.__name__}` decorator on the subclass'
            )
        super().__init__()
        self._event_handlers = self.EVENT_HANDLERS_BY_CLASS[self.__class__]
        self.tracing_enabled = False
        self.tracing_reset_pending = False
        self.sys_tracer = self._sys_tracer
        self.existing_tracer = None

    def _emit_event(self, evt: Union[TraceEvent, str], node_id: int, **kwargs: Any):
        event = TraceEvent(evt) if isinstance(evt, str) else evt
        frame = kwargs.get('_frame', sys._getframe().f_back)
        kwargs['_frame'] = frame
        for handler in self._event_handlers[event]:
            try:
                new_ret = handler(self, kwargs.get('ret', None), node_id, frame, event, **kwargs)
            except Exception as exc:
                if SafetyRunMode.get() == SafetyRunMode.DEVELOP:
                    raise exc
                else:
                    logger.error('Exception occurred: %s', str(exc))
                new_ret = None
            if new_ret is not None:
                kwargs['ret'] = new_ret
        return kwargs.get('ret', None)

    def _make_stack(self):
        return TraceStack(self)

    def _make_composed_tracer(self, existing_tracer):  # pragma: no cover

        @functools.wraps(self._sys_tracer)
        def _composed_tracer(frame: FrameType, evt: str, arg: Any, **kwargs):
            existing_ret = existing_tracer(frame, evt, arg, **kwargs)
            if not self.tracing_enabled:
                return existing_ret
            my_ret = self._sys_tracer(frame, evt, arg, **kwargs)
            if my_ret is None and evt == 'call':
                return existing_ret
            else:
                return my_ret
        return _composed_tracer

    def _settrace_patch(self, trace_func):  # pragma: no cover
        # called by third-party tracers
        self.existing_tracer = trace_func
        if self.tracing_enabled:
            if trace_func is None:
                self._disable_tracing()
            self._enable_tracing(check_disabled=False, existing_tracer=trace_func)
        else:
            nbs().settrace(trace_func)

    def _enable_tracing(self, check_disabled=True, existing_tracer=None):
        if check_disabled:
            assert not self.tracing_enabled
        self.tracing_enabled = True
        self.existing_tracer = existing_tracer or sys.gettrace()
        if self.existing_tracer is None:
            self.sys_tracer = self._sys_tracer
        else:
            self.sys_tracer = self._make_composed_tracer(self.existing_tracer)
        nbs().settrace(self.sys_tracer)

    def _disable_tracing(self, check_enabled=True):
        if check_enabled:
            assert self.tracing_enabled
            assert sys.gettrace() is self.sys_tracer
        self.tracing_enabled = False
        nbs().settrace(self.existing_tracer)

    @contextmanager
    def _patch_sys_settrace(self):
        original_settrace = sys.settrace
        try:
            sys.settrace = self._settrace_patch
            yield
        finally:
            sys.settrace = original_settrace

    @contextmanager
    def tracing_context(self):
        setattr(builtins, EMIT_EVENT, self._emit_event)
        try:
            with self._patch_sys_settrace():
                self._enable_tracing()
                yield
        finally:
            delattr(builtins, EMIT_EVENT)
            self._disable_tracing(check_enabled=False)

    def _attempt_to_reenable_tracing(self, frame: FrameType):
        return NotImplemented

    def _sys_tracer(self, frame: FrameType, evt: str, arg: Any, **__):
        if self.tracing_reset_pending:
            assert evt == 'call', 'expected call; got event %s' % evt
            self._attempt_to_reenable_tracing(frame)
            return None
        if evt == 'line' or not nbs().is_cell_file(frame.f_code.co_filename):
            return None

        return self._emit_event(evt, 0, _frame=frame, ret=arg)


def register_handler(event: Union[TraceEvent, Tuple[TraceEvent, ...]]):
    events = event if isinstance(event, tuple) else (event,)

    def _inner_registrar(handler):
        for evt in events:
            BaseTraceManager.EVENT_HANDLERS_PENDING_REGISTRATION[evt].append(handler)
        return handler
    return _inner_registrar


def register_universal_handler(handler):
    return register_handler(tuple(evt for evt in TraceEvent))(handler)


def register_trace_manager_class(mgr_cls: Type[BaseTraceManager]) -> Type[BaseTraceManager]:
    mgr_cls.EVENT_HANDLERS_BY_CLASS[mgr_cls] = defaultdict(list, mgr_cls.EVENT_HANDLERS_PENDING_REGISTRATION)
    mgr_cls.EVENT_HANDLERS_PENDING_REGISTRATION.clear()
    mgr_cls._MANAGER_CLASS_REGISTERED = True
    return mgr_cls


@register_trace_manager_class
class TraceManager(BaseTraceManager):
    def __init__(self):
        super().__init__()
        self.trace_event_counter = 0
        self.prev_event: Optional[TraceEvent] = None
        self.prev_trace_stmt: Optional[TraceStatement] = None
        self.seen_stmts: Set[NodeId] = set()
        self.call_depth = 0
        self.traced_statements: Dict[NodeId, TraceStatement] = {}
        self.node_id_to_loaded_symbols: Dict[NodeId, List[DataSymbol]] = defaultdict(list)
        self.node_id_to_saved_store_data: Dict[NodeId, SavedStoreData] = {}
        self.node_id_to_saved_del_data: Dict[NodeId, SavedDelData] = {}
        self.node_id_to_loaded_literal_scope: Dict[NodeId, NamespaceScope] = {}
        self.node_id_to_saved_dict_key: Dict[NodeId, Any] = {}

        self.call_stack: TraceStack = self._make_stack()
        with self.call_stack.register_stack_state():
            # everything here should be copyable
            self.prev_trace_stmt_in_cur_frame: Optional[TraceStatement] = None
            self.prev_node_id_in_cur_frame: Optional[NodeId] = None
            self.mutations: List[Mutation] = []
            self.mutation_candidates: List[MutationCandidate] = []
            self.saved_assign_rhs_obj_id: Optional[int] = None

            with self.call_stack.needing_manual_initialization():
                self.cur_frame_original_scope: Scope = nbs().global_scope
                self.active_scope: Scope = nbs().global_scope
                self.inside_anonymous_call = False

            self.lexical_call_stack: TraceStack = self._make_stack()
            with self.lexical_call_stack.register_stack_state():
                self.num_args_seen = 0
                self.sym_for_obj_calling_method: Optional[DataSymbol] = None
                self.first_obj_id_in_chain: Optional[ObjId] = None
                self.top_level_node_id_for_chain: Optional[NodeId] = None
                self.saved_complex_symbol_load_data: Optional[SavedComplexSymbolLoadData] = None
                self.prev_node_id_in_cur_frame_lexical: Optional[NodeId] = None

            self.lexical_literal_stack: TraceStack = self._make_stack()
            with self.lexical_literal_stack.register_stack_state():
                # `None` means use 'cur_frame_original_scope'
                self.active_literal_scope: Optional[NamespaceScope] = None

    # TODO: use stack mechanism to automate this?
    def after_stmt_reset_hook(self):
        self.mutations.clear()
        self.mutation_candidates.clear()
        self.lexical_call_stack.clear()
        self.lexical_literal_stack.clear()
        self.active_scope = self.cur_frame_original_scope
        self.first_obj_id_in_chain = None
        self.top_level_node_id_for_chain = None
        self.saved_complex_symbol_load_data = None
        self.active_literal_scope = None
        self.node_id_to_loaded_literal_scope.clear()
        self.node_id_to_saved_dict_key.clear()

    def _handle_call_transition(self, trace_stmt: TraceStatement):
        # ensures we only handle del's and not delitem's
        self.node_id_to_saved_del_data.clear()
        new_scope = trace_stmt.get_post_call_scope()
        with self.call_stack.push():
            # TODO: figure out a better way to determine if we're inside a lambda
            #  could this one lead to a false negative if a lambda is in the default of a function def kwarg?
            self.inside_anonymous_call = not isinstance(
                trace_stmt.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
            )
            self.cur_frame_original_scope = new_scope
            self.active_scope = new_scope
        self.prev_trace_stmt_in_cur_frame = self.prev_trace_stmt = trace_stmt

    def _check_prev_stmt_done_executing_hook(self, event: TraceEvent, trace_stmt: TraceStatement):
        if event == TraceEvent.after_stmt:
            trace_stmt.finished_execution_hook()
        elif event == TraceEvent.return_ and self.prev_event not in (TraceEvent.call, TraceEvent.exception):
            # ensuring prev != call ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            if self.prev_trace_stmt is not None:
                self.prev_trace_stmt.finished_execution_hook()
            # prev_overall = self.prev_trace_stmt
            # if prev_overall is not None and prev_overall is not self._stack[-1][0]:
            #     # this condition ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            #     prev_overall.finished_execution_hook()

    def _handle_return_transition(self, trace_stmt: TraceStatement, ret: Any):
        try:
            inside_anonymous_call = self.inside_anonymous_call
            return_to_stmt: TraceStatement = self.call_stack.get_field('prev_trace_stmt_in_cur_frame')
            assert return_to_stmt is not None
            if self.prev_event != TraceEvent.exception:
                # exception events are followed by return events until we hit an except clause
                # no need to track dependencies in this case
                if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
                    return_to_stmt.class_scope = cast(NamespaceScope, self.cur_frame_original_scope)
                elif isinstance(trace_stmt.stmt_node, ast.Return) or inside_anonymous_call:
                    if not trace_stmt.lambda_call_point_deps_done_once:
                        trace_stmt.lambda_call_point_deps_done_once = True
                        maybe_lambda_sym = nbs().statement_to_func_cell.get(id(trace_stmt.stmt_node), None)
                        maybe_lambda_node = None
                        if maybe_lambda_sym is not None:
                            maybe_lambda_node = maybe_lambda_sym.stmt_node
                        if inside_anonymous_call and maybe_lambda_node is not None and isinstance(maybe_lambda_node, ast.Lambda):
                            rvals = resolve_rval_symbols(maybe_lambda_node.body)
                        else:
                            rvals = resolve_rval_symbols(trace_stmt.stmt_node)
                        dsym_to_attach = None
                        if len(rvals) == 1:
                            dsym_to_attach = next(iter(rvals))
                            if dsym_to_attach.obj_id != id(ret):
                                dsym_to_attach = None
                        if dsym_to_attach is None and len(rvals) > 0:
                            dsym_to_attach = self.cur_frame_original_scope.upsert_data_symbol_for_name(
                                '<return_sym_%d>' % id(ret), ret, rvals, trace_stmt.stmt_node, is_anonymous=True
                            )
                        if dsym_to_attach is not None:
                            return_to_node_id = self.call_stack.get_field('prev_node_id_in_cur_frame')
                            # logger.error("prev seen: %s", ast.dump(nbs().ast_node_by_id[return_to_node_id]))
                            try:
                                call_node_id = self.call_stack.get_field(
                                    'lexical_call_stack'
                                ).get_field('prev_node_id_in_cur_frame_lexical')
                                call_node = cast(ast.Call, nbs().ast_node_by_id[call_node_id])
                                # logger.error("prev seen outer: %s", ast.dump(nbs().ast_node_by_id[call_node_id]))
                                total_args = len(call_node.args) + len(call_node.keywords)
                                num_args_seen = self.call_stack.get_field('num_args_seen')
                                logger.warning("num args seen: %d", num_args_seen)
                                if total_args == num_args_seen:
                                    return_to_node_id = call_node_id
                                else:
                                    assert num_args_seen < total_args
                                    if num_args_seen < len(call_node.args):
                                        return_to_node_id = id(call_node.args[num_args_seen])
                                    else:
                                        return_to_node_id = id(call_node.keywords[num_args_seen - len(call_node.args)].value)
                            except IndexError:
                                pass
                            # logger.error("use node %s", ast.dump(nbs().ast_node_by_id[return_to_node_id]))
                            self.node_id_to_loaded_symbols[return_to_node_id].append(dsym_to_attach)
        finally:
            self.call_stack.pop()

    def state_transition_hook(
        self,
        event: TraceEvent,
        trace_stmt: TraceStatement,
        ret: Any,
    ):
        self.trace_event_counter += 1

        self._check_prev_stmt_done_executing_hook(event, trace_stmt)

        if event == TraceEvent.call:
            self._handle_call_transition(trace_stmt)
        if event == TraceEvent.return_:
            self._handle_return_transition(trace_stmt, ret)
        self.prev_event = event

    @staticmethod
    def _partial_resolve_ref(ref: Union[str, int, ast.AST]) -> Union[str, int]:
        if isinstance(ref, ast.Starred):
            ref = ref.value
        if isinstance(ref, ast.Name):
            ref = ref.id
        if isinstance(ref, ast.AST):
            ref = id(ref)
        return ref

    def resolve_store_data_for_target(
        self, target: Union[str, int, ast.AST], frame: FrameType
    ) -> Tuple[Scope, AttrSubVal, Any, bool]:
        target = self._partial_resolve_ref(target)
        if isinstance(target, str):
            obj = frame.f_locals[target]
            return self.cur_frame_original_scope, target, obj, False
        (
            scope, obj, attr_or_sub, is_subscript
        ) = self.node_id_to_saved_store_data[target]
        attr_or_sub_obj = nbs().retrieve_namespace_attr_or_sub(obj, attr_or_sub, is_subscript)
        if attr_or_sub_obj is None:
            scope_to_use = scope
        else:
            scope_to_use = scope.get_earliest_ancestor_containing(id(attr_or_sub_obj), is_subscript)
        if scope_to_use is None:
            # Nobody before `scope` has it, so we'll insert it at this level
            scope_to_use = scope
        return scope_to_use, attr_or_sub, attr_or_sub_obj, is_subscript

    def resolve_del_data_for_target(self, target: Union[str, int, ast.AST]) -> Tuple[Scope, AttrSubVal, Any, bool]:
        target = self._partial_resolve_ref(target)
        if isinstance(target, str):
            return self.cur_frame_original_scope, target, None, False
        (
            scope, attr_or_sub, is_subscript
        ) = self.node_id_to_saved_del_data[target]
        return scope, attr_or_sub, None, is_subscript

    def resolve_loaded_symbols(self, symbol_ref: Union[str, int, ast.AST, DataSymbol]) -> List[DataSymbol]:
        if isinstance(symbol_ref, DataSymbol):
            return [symbol_ref]
        symbol_ref = self._partial_resolve_ref(symbol_ref)
        if isinstance(symbol_ref, int):
            return self.node_id_to_loaded_symbols.get(symbol_ref, [])
        elif isinstance(symbol_ref, str):
            return [self.cur_frame_original_scope.lookup_data_symbol_by_name(symbol_ref)]
        else:
            return []

    def resolve_symbols(self, symbol_refs: Set[Union[str, int, DataSymbol]]) -> Set[DataSymbol]:
        data_symbols = set()
        for ref in symbol_refs:
            data_symbols.update(self.resolve_loaded_symbols(ref))
        return data_symbols

    def _get_namespace_for_obj(self, obj: Any, obj_name: Optional[str] = None) -> NamespaceScope:
        obj_id = id(obj)
        ns = nbs().namespaces.get(obj_id, None)
        if ns is not None:
            return ns
        class_scope = nbs().namespaces.get(id(obj.__class__), None)
        if class_scope is not None:
            # logger.warning(
            #     'found class scope %s containing %s',
            #     class_scope, list(class_scope.all_data_symbols_this_indentation())
            # )
            ns = class_scope.clone(obj)
            if obj_name is not None:
                ns.scope_name = obj_name
        else:
            # print('no scope for class', obj.__class__)
            try:
                scope_name = nbs().get_first_full_symbol(obj_id).name if obj_name is None else obj_name
            except AttributeError:
                scope_name = '<unknown namespace>'
            ns = NamespaceScope(obj, scope_name, parent_scope=None)
        # FIXME: brittle strategy for determining parent scope of obj
        if ns.parent_scope is None:
            if (
                obj_name is not None and
                obj_name not in self.prev_trace_stmt_in_cur_frame.frame.f_locals
            ):
                parent_scope = nbs().global_scope
            else:
                parent_scope = self.active_scope
            ns.parent_scope = parent_scope
        return ns

    def _clear_info_and_maybe_lookup_or_create_complex_symbol(self, obj_attr_or_sub) -> Optional[DataSymbol]:
        if self.saved_complex_symbol_load_data is None:
            return None
        (scope, obj), (attr_or_subscript, is_subscript) = self.saved_complex_symbol_load_data
        self.saved_complex_symbol_load_data = None
        data_sym = scope.lookup_data_symbol_by_name_this_indentation(
            attr_or_subscript, is_subscript=is_subscript, skip_cloned_lookup=True,
        )
        logger.warning("found sym %s in scope %s", data_sym, scope)
        if data_sym is None:
            parent = scope.lookup_data_symbol_by_name_this_indentation(
                attr_or_subscript, is_subscript, skip_cloned_lookup=False,
            )
            parents = set() if parent is None else {parent}
            is_default_dict = isinstance(obj, defaultdict)
            data_sym = scope.upsert_data_symbol_for_name(
                attr_or_subscript,
                obj_attr_or_sub,
                parents,
                self.prev_trace_stmt_in_cur_frame.stmt_node,
                is_subscript=is_subscript,
                propagate=is_default_dict,
                implicit=not is_default_dict,
            )
            logger.warning("create implicit sym %s", data_sym)
        elif data_sym.obj_id != id(obj_attr_or_sub):
            data_sym.update_obj_ref(obj_attr_or_sub)
        return data_sym

    @register_handler(
        # all the AST-related events
        tuple(set(TraceEvent) - {
            TraceEvent.call,
            TraceEvent.return_,
            TraceEvent.exception,
            TraceEvent.c_call,
            TraceEvent.c_return,
            TraceEvent.c_exception,
            TraceEvent.argument,
        })
    )
    def _save_node_id(self, _obj, node_id: NodeId, *_, **__):
        self.prev_node_id_in_cur_frame = node_id
        self.prev_node_id_in_cur_frame_lexical = node_id

    @register_handler(TraceEvent.init_cell)
    def init_cell(self, _obj, _node_id, frame: FrameType, _event, cell_id: Union[str, int], **__):
        nbs().set_name_to_cell_num_mapping(frame)

    @register_handler(TraceEvent.after_assign_rhs)
    def after_assign_rhs(self, obj: Any, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        self.saved_assign_rhs_obj_id = id(obj)

    @register_handler((TraceEvent.attribute, TraceEvent.subscript))
    def attrsub_tracer(
        self,
        obj: Any,
        node_id: NodeId,
        _frame_: FrameType,
        event: TraceEvent,
        *_,
        attr_or_subscript: AttrSubVal,
        ctx: str,
        call_context: bool,
        top_level_node_id: NodeId,
        obj_name: Optional[str] = None,
        **__
    ):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        if isinstance(nbs().ast_node_by_id[node_id], ast.Call):
            # clear the callpoint dependency
            self.node_id_to_loaded_symbols.pop(node_id, None)
        if obj is None or obj is get_ipython():
            return
        logger.warning('%s attrsub %s of obj %s', ctx, attr_or_subscript, obj)
        sym_for_obj = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
        
        # Resolve symbol if necessary
        if sym_for_obj is None and obj_name is not None:
            sym_for_obj = self.active_scope.lookup_data_symbol_by_name_this_indentation(obj_name)

        if sym_for_obj is not None and sym_for_obj.timestamp < nbs().cell_counter():
            sym_for_obj.timestamp_by_used_time[nbs().cell_counter()] = sym_for_obj.timestamp_excluding_ns_descendents
        
        is_subscript = (event == TraceEvent.subscript)

        obj_id = id(obj)
        if self.top_level_node_id_for_chain is None:
            self.top_level_node_id_for_chain = top_level_node_id
        if self.first_obj_id_in_chain is None:
            self.first_obj_id_in_chain = obj_id
        if isinstance(attr_or_subscript, tuple):
            if not all(isinstance(v, (str, int)) for v in attr_or_subscript):
                return
        elif not isinstance(attr_or_subscript, (str, int)):
            return

        scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
        self.active_scope = scope

        if ctx in ('Store', 'AugStore'):
            logger.warning(
                "save store data for node id %d: %s, %s, %s, %s",
                top_level_node_id, scope, obj, attr_or_subscript, is_subscript
            )
            self.node_id_to_saved_store_data[top_level_node_id] = (scope, obj, attr_or_subscript, is_subscript)
            return
        elif ctx == 'Del':
            # logger.error("save del data for node %s", ast.dump(nbs().ast_node_by_id[top_level_node_id]))
            logger.warning("save del data for node id %d", top_level_node_id)
            self.node_id_to_saved_del_data[top_level_node_id] = (scope, attr_or_subscript, is_subscript)
            return
        if call_context:
            mutation_event: MutationEvent = StandardMutation()
            if isinstance(obj, list):
                if attr_or_subscript == 'append':
                    mutation_event = ListAppend()
                elif attr_or_subscript == 'extend':
                    mutation_event = ListExtend(len(obj))
                elif attr_or_subscript == 'insert':
                    mutation_event = ListInsert()
            # save off event counter and obj_id
            # if event counter didn't change when we process the Call retval, and if the
            # retval is None, this is a likely signal that we have a mutation
            self.mutation_candidates.append(
                ((self.trace_event_counter, obj_id, obj_name), mutation_event, set(), [])
            )
            if not is_subscript:
                if sym_for_obj is None and obj_name is not None:
                    sym_for_obj = self.cur_frame_original_scope.lookup_data_symbol_by_name(
                        obj_name, is_subscript=is_subscript
                    )
                if sym_for_obj is None and self.prev_trace_stmt_in_cur_frame is not None:
                    sym_for_obj = self.cur_frame_original_scope.upsert_data_symbol_for_name(
                        obj_name or '<anonymous_symbol_%d>' % id(obj),
                        obj,
                        set(),
                        self.prev_trace_stmt_in_cur_frame.stmt_node,
                        is_subscript=is_subscript,
                        is_anonymous=obj_name is None,
                        propagate=False,
                        implicit=True,
                    )
                if sym_for_obj is not None:
                    self.sym_for_obj_calling_method = sym_for_obj
        else:
            logger.warning("saved load data: %s, %s, %s", scope, attr_or_subscript, is_subscript)
            self.saved_complex_symbol_load_data = ((scope, obj), (attr_or_subscript, is_subscript))

    @register_handler(TraceEvent.after_complex_symbol)
    def after_complex_symbol(self, obj: Any, *_, call_context: bool, ctx: str, **__):
        try:
            if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
                return
            if self.first_obj_id_in_chain is None:
                return
            if ctx != 'Load':
                # don't trace after non-load events
                return
            loaded_sym = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
            if call_context and len(self.mutation_candidates) > 0:
                (
                    (evt_counter, obj_id, obj_name),
                    mutation_event,
                    recorded_arg_dsyms,
                    recorded_arg_objs,
                ) = self.mutation_candidates.pop()
                if evt_counter == self.trace_event_counter:
                    if obj is None or id(obj) == obj_id:
                        if isinstance(mutation_event, StandardMutation):
                            try:
                                top_level_sym = nbs().get_first_full_symbol(self.first_obj_id_in_chain)
                                if top_level_sym.is_import and top_level_sym.name not in ARG_MUTATION_EXCEPTED_MODULES:
                                    # TODO: should it be the other way around?
                                    #  i.e. allow-list for arg mutations, starting with np.random.seed?
                                    if len(recorded_arg_dsyms) > 0:
                                        # only make this an arg mutation event if it looks like there's an arg to mutate
                                        mutation_event = ArgMutate()
                            except:
                                pass
                        elif isinstance(mutation_event, ListInsert):
                            mutation_event.insert_pos = recorded_arg_objs[0]
                        self.mutations.append((obj_id, mutation_event, recorded_arg_dsyms, recorded_arg_objs))
                    else:
                        if self.sym_for_obj_calling_method is not None:
                            loaded_sym = self.sym_for_obj_calling_method
                            self.sym_for_obj_calling_method = None
            if loaded_sym is not None:
                self.node_id_to_loaded_symbols[self.top_level_node_id_for_chain].append(loaded_sym)
        finally:
            self.saved_complex_symbol_load_data = None
            self.first_obj_id_in_chain = None
            self.top_level_node_id_for_chain = None
            self.sym_for_obj_calling_method = None
            self.active_scope = self.cur_frame_original_scope

    @register_handler(TraceEvent.argument)
    def argument(self, arg_obj: Any, arg_node_id: int, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        self.num_args_seen += 1
        arg_node = nbs().ast_node_by_id.get(arg_node_id, None)
        if len(self.mutation_candidates) == 0:
            return

        if isinstance(arg_node, ast.Name):
            assert self.active_scope is self.cur_frame_original_scope
            arg_dsym = self.active_scope.lookup_data_symbol_by_name(arg_node.id)
            if arg_dsym is None:
                self.active_scope.upsert_data_symbol_for_name(
                    arg_node.id, arg_obj, set(), self.prev_trace_stmt_in_cur_frame.stmt_node, implicit=True
                )
        self.mutation_candidates[-1][-2].update(resolve_rval_symbols(arg_node))
        self.mutation_candidates[-1][-1].append(arg_obj)

    @register_handler(TraceEvent.before_call)
    def before_call(self, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        with self.lexical_call_stack.push():
            pass
        self.active_scope = self.cur_frame_original_scope

    @register_handler(TraceEvent.after_call)
    def after_call(self, *_, call_node_id: NodeId, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        # no need to reset active scope here;
        # that will happen in the 'after chain' handler
        self.lexical_call_stack.pop()

    # Note: we don't trace set literals
    @register_handler((TraceEvent.before_dict_literal, TraceEvent.before_list_literal, TraceEvent.before_tuple_literal))
    def before_literal(self, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        parent_scope = self.active_literal_scope or self.cur_frame_original_scope
        with self.lexical_literal_stack.push():
            self.active_literal_scope = NamespaceScope(None, NamespaceScope.ANONYMOUS, parent_scope)

    @register_handler((TraceEvent.after_dict_literal, TraceEvent.after_list_literal, TraceEvent.after_tuple_literal))
    def after_literal(self, literal: Union[dict, list, tuple], node_id: NodeId, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return literal
        try:
            self.active_literal_scope.update_obj_ref(literal)
            logger.warning("create literal scope %s", self.active_literal_scope)
            starred_idx = -1
            starred_namespace = None
            for (i, inner_obj), (inner_key_node, inner_val_node) in match_container_obj_or_namespace_with_literal_nodes(
                literal, nbs().ast_node_by_id[node_id]  # type: ignore
            ):
                # TODO: memoize symbol resolution; otherwise this will be quadratic for deeply nested literals
                if isinstance(inner_val_node, ast.Starred):
                    inner_symbols = set()
                    starred_idx += 1
                    if starred_idx == 0:
                        starred_syms = self.resolve_loaded_symbols(inner_val_node)
                        starred_namespace = nbs().namespaces.get(starred_syms[0].obj_id, None) if starred_syms else None
                    if starred_namespace is not None:
                        starred_dep = starred_namespace.lookup_data_symbol_by_name_this_indentation(starred_idx, is_subscript=True)
                        inner_symbols.add(starred_dep)
                else:
                    inner_symbols = resolve_rval_symbols(inner_val_node)
                    if inner_key_node is not None:
                        # inner_symbols.add(self.resolve_loaded_symbols(inner_key_node))
                        inner_symbols.update(resolve_rval_symbols(inner_key_node))
                self.node_id_to_loaded_symbols.pop(id(inner_val_node), None)
                inner_symbols.discard(None)
                if isinstance(i, (int, str)):  # TODO: perform more general check for SupportedIndexType
                    self.active_literal_scope.upsert_data_symbol_for_name(
                        i, inner_obj, inner_symbols, self.prev_trace_stmt_in_cur_frame.stmt_node, is_subscript=True
                    )
            self.node_id_to_loaded_literal_scope[node_id] = self.active_literal_scope
            parent_scope: Scope = self.active_literal_scope.parent_scope
            assert parent_scope is not None
            literal_sym = parent_scope.upsert_data_symbol_for_name(
                '<literal_sym_%d>' % id(literal),
                literal,
                set(),
                self.prev_trace_stmt_in_cur_frame.stmt_node,
                is_anonymous=True,
            )
            self.node_id_to_loaded_symbols[node_id].append(literal_sym)
            return literal
        finally:
            self.lexical_literal_stack.pop()

    @register_handler(TraceEvent.dict_key)
    def dict_key(self, obj: Any, key_node_id: NodeId, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return obj
        self.node_id_to_saved_dict_key[key_node_id] = obj
        return obj

    @register_handler(TraceEvent.dict_value)
    def dict_value(self, obj: Any, value_node_id: NodeId, *_, key_node_id: NodeId, dict_node_id: NodeId, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return obj
        scope = self.node_id_to_loaded_literal_scope.pop(value_node_id, None)
        if scope is None:
            return obj
        # if we found a pending literal, assert that it's not dict unpacking
        assert key_node_id is not None
        key_obj = self.node_id_to_saved_dict_key.pop(key_node_id, None)
        if isinstance(key_obj, (str, int)):
            scope.scope_name = str(key_obj)
        return obj

    @register_handler((TraceEvent.list_elt, TraceEvent.tuple_elt))
    def list_or_tuple_elt(
        self, obj: Any, elt_node_id: NodeId, *_, index: Optional[int], container_node_id: NodeId, **__
    ):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return obj
        scope = self.node_id_to_loaded_literal_scope.pop(elt_node_id, None)
        if scope is None:
            return obj
        if index is not None:
            scope.scope_name = str(index)
        return obj

    @register_handler(TraceEvent.after_lambda)
    def after_lambda(self, obj: Any, lambda_node_id: int, frame: FrameType, *_, **__):
        sym_deps = []
        node = nbs().ast_node_by_id[lambda_node_id]
        for kw_default in node.args.defaults:  # type: ignore
            sym_deps.extend(self.resolve_loaded_symbols(kw_default))
        sym = self.active_scope.upsert_data_symbol_for_name(
            '<lambda_sym_%d>' % id(obj),
            obj,
            sym_deps,
            self.prev_trace_stmt_in_cur_frame.stmt_node,
            is_function_def=True,
            propagate=False,
        )
        # FIXME: this is super brittle. We're passing in a stmt node to update the mapping from
        #  stmt_node to function symbol, but simultaneously forcing the lambda symbol to hold
        #  a reference to the lambda in order to help with symbol resolution later
        sym.stmt_node = node
        self.node_id_to_loaded_symbols[lambda_node_id].append(sym)

    @register_handler(TraceEvent.after_stmt)
    def after_stmt(self, ret_expr: Any, stmt_id: int, frame: FrameType, *_, **__):
        if stmt_id in self.seen_stmts:
            return ret_expr
        stmt = nbs().ast_node_by_id.get(stmt_id, None)
        if stmt is not None:
            self.handle_sys_events(None, 0, frame, TraceEvent.after_stmt, stmt_node=cast(ast.stmt, stmt))
        return ret_expr

    @register_handler(TraceEvent.before_stmt)
    def before_stmt(self, _ret: None, stmt_id: int, frame: FrameType, *_, **__) -> None:
        if stmt_id in self.seen_stmts:
            return
        # logger.warning('reenable tracing: %s', site_id)
        if self.prev_trace_stmt_in_cur_frame is not None:
            prev_trace_stmt_in_cur_frame = self.prev_trace_stmt_in_cur_frame
            # both of the following stmts should be processed when body is entered
            if isinstance(prev_trace_stmt_in_cur_frame.stmt_node, (ast.For, ast.If, ast.With)):
                self.after_stmt(None, prev_trace_stmt_in_cur_frame.stmt_id, frame)
        trace_stmt = self.traced_statements.get(stmt_id, None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(frame, cast(ast.stmt, nbs().ast_node_by_id[stmt_id]))
            self.traced_statements[stmt_id] = trace_stmt
        self.prev_trace_stmt_in_cur_frame = trace_stmt
        self.prev_trace_stmt = trace_stmt
        if not self.tracing_enabled:
            assert not self.tracing_reset_pending
            self._enable_tracing()
            self.tracing_reset_pending = True
            _finish_tracing_reset()  # trigger the tracer with a frame

    def _attempt_to_reenable_tracing(self, frame: FrameType) -> None:
        if nbs().is_develop:
            assert self.tracing_reset_pending, 'expected tracing reset to be pending!'
            assert self.call_depth > 0, 'expected managed call depth > 0, got %d' % self.call_depth
        self.tracing_reset_pending = False
        call_depth = 0
        while frame is not None:
            if nbs().is_cell_file(frame.f_code.co_filename):
                call_depth += 1
            frame = frame.f_back
        if nbs().is_develop:
            assert call_depth >= 1, 'expected call depth >= 1, got %d' % call_depth
        # TODO: allow reenabling tracing beyond just at the top level
        if call_depth != 1:
            self._disable_tracing()
            return
        # at this point, we can be sure we're at the top level
        # because tracing was enabled in a handler and not in the
        # top level, we need to clear the stack, since we won't
        # catch the return event
        self.call_depth = 0
        self.call_stack.clear()
        self.lexical_call_stack.clear()
        if nbs().trace_messages_enabled:
            self.EVENT_LOGGER.warning('reenable tracing >>>')

    @register_handler((TraceEvent.call, TraceEvent.return_, TraceEvent.exception))
    def handle_sys_events(
        self,
        ret_obj: Any,
        _node_id: int,
        frame: FrameType,
        event: TraceEvent,
        *_,
        stmt_node: Optional[ast.stmt] = None,
        **__
    ):
        # right now, this should only be enabled for notebook code
        assert nbs().is_cell_file(frame.f_code.co_filename), 'got %s' % frame.f_code.co_filename
        assert self.tracing_enabled or event == TraceEvent.after_stmt

        # IPython quirk -- every line in outer scope apparently wrapped in lambda
        # We want to skip the outer 'call' and 'return' for these
        if event == TraceEvent.call:
            self.call_depth += 1
            if self.call_depth == 1:
                return self.sys_tracer

        if event == TraceEvent.return_:
            self.call_depth -= 1
            if self.call_depth == 0:
                return

        cell_num, lineno = nbs().get_position(frame)
        if cell_num is None:
            return None

        if event == TraceEvent.after_stmt:
            assert stmt_node is not None
        else:
            try:
                stmt_node = nbs().statement_cache[cell_num][lineno]
                if event == TraceEvent.call and not isinstance(stmt_node, (ast.AsyncFunctionDef, ast.FunctionDef)):
                    # TODO: this is bad and I should feel bad. Need a better way to figure out which
                    #  stmt is executing than by using line numbers.
                    parent_node = nbs().parent_node_by_id.get(id(stmt_node), None)
                    logger.error("node %s parent %s", ast.dump(stmt_node), None if parent_node is None else ast.dump(parent_node))
                    if (
                        parent_node is not None
                        and getattr(parent_node, 'lineno', None) == lineno
                        and isinstance(parent_node, (ast.AsyncFunctionDef, ast.FunctionDef))
                    ):
                        stmt_node = parent_node
            except KeyError as e:
                self.EVENT_LOGGER.warning("got key error for stmt node in cell %d, line %d", cell_num, lineno)
                if nbs().is_develop:
                    raise e
                return self.sys_tracer

        trace_stmt = self.traced_statements.get(id(stmt_node), None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(frame, stmt_node)
            self.traced_statements[id(stmt_node)] = trace_stmt

        if nbs().trace_messages_enabled:
            codeline = astunparse.unparse(stmt_node).strip('\n').split('\n')[0]
            codeline = ' ' * getattr(stmt_node, 'col_offset', 0) + codeline
            self.EVENT_LOGGER.warning(' %3d: %10s >>> %s', trace_stmt.lineno, event, codeline)
        if event == TraceEvent.call:
            if trace_stmt.node_id_for_last_call == self.prev_node_id_in_cur_frame:
                if nbs().trace_messages_enabled:
                    self.EVENT_LOGGER.warning(' disable tracing >>>')
                self._disable_tracing()
                return None
            trace_stmt.node_id_for_last_call = self.prev_node_id_in_cur_frame
        self.state_transition_hook(event, trace_stmt, ret_obj)
        return self.sys_tracer


assert not BaseTraceManager._MANAGER_CLASS_REGISTERED
assert TraceManager._MANAGER_CLASS_REGISTERED



# -*- coding: future_annotations -*-
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


class MutationEvent:
    pass


class StandardMutation(MutationEvent):
    pass


class ListAppend(MutationEvent):
    pass


class ListExtend(MutationEvent):
    def __init__(self, orig_len: int) -> None:
        super().__init__()
        self.orig_len: int = orig_len


class ListInsert(MutationEvent):
    def __init__(self, insert_pos: Optional[int] = None):
        super().__init__()
        self.insert_pos = insert_pos


class ArgMutate(MutationEvent):
    pass




# -*- coding: future_annotations -*-
import typing
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, List, Iterable

_PRIMITIVE_TYPES = {int, float, str, type(None)}
_CONTAINER_TYPES = {
    dict: typing.Dict,
    list: typing.List,
    set: typing.Set,
    tuple: typing.Tuple,
}


def _get_contained_type_annotations(container: Iterable[Any]) -> List:
    annotations = []
    for obj in container:
        annotations.append(get_type_annotation(obj))
    return annotations


def _resolve_container_types(container_type, types: list):
    used_types = set(types)
    none_seen = type(None) in used_types
    used_types.discard(type(None))
    if float in used_types:
        used_types.discard(int)
    if len(used_types) == 0:
        return container_type[None] if none_seen else container_type
    elif len(used_types) == 1:
        ret = next(iter(used_types))
    else:
        ret = typing.Union[tuple(used_types)]
    if none_seen:
        ret = typing.Optional[ret]
    return container_type[ret]


def _resolve_tuple_types(types: list):
    if len(types) == 0:
        return typing.Tuple
    elif len(types) <= 3:
        return typing.Tuple[tuple(types)]
    else:
        args = _resolve_container_types(typing.List, types).__args__
        if len(args) == 1:
            return typing.Tuple[args[0], ...]
        else:
            return typing.Tuple[tuple(types)]


def get_type_annotation(obj):
    obj_type = type(obj)
    if obj_type in _PRIMITIVE_TYPES:
        return obj_type
    elif obj_type in _CONTAINER_TYPES:
        if obj_type == dict:
            key_ann = _resolve_container_types(typing.List, _get_contained_type_annotations(obj.keys()))
            value_ann = _resolve_container_types(typing.List, _get_contained_type_annotations(obj.values()))
            key_args = getattr(key_ann, '__args__', None)
            value_args = getattr(value_ann, '__args__', None)
            if key_args is None or value_args is None:
                return typing.Dict
            else:
                if len(key_args) == 1:
                    key_args = key_args[0]
                if len(value_args) == 1:
                    value_args = value_args[0]
                return typing.Dict[key_args, value_args]
        elif obj_type == tuple:
            return _resolve_tuple_types(_get_contained_type_annotations(obj))
        else:
            return _resolve_container_types(
                _CONTAINER_TYPES[obj_type], _get_contained_type_annotations(obj)
            )
    elif obj_type is type:
        return typing.Type[obj]
    else:
        return obj_type


def make_annotation_string(ann) -> str:
    if ann is type(None):
        ret = 'None'
    elif hasattr(ann, '__name__'):
        ret = ann.__name__
    elif hasattr(ann, '_name'):
        ret = ann._name
        if ret is None:
            args = ann.__args__
            if args[-1] is type(None) and len(args) == 2:
                ret = 'Optional'
            else:
                ret = 'Union'
    elif ann is ...:
        ret = '...'
    else:
        ret = str(ann)

    if ret.startswith('typing.') and '[' in ret:
        ret = ret.split('.')[1].split('[')[0]

    module = getattr(ann, '__module__', None)
    if module is not None and module not in ('typing', 'builtins', '__main__'):
        ret = f'{module}.{ret}'

    ann_args = getattr(ann, '__args__', None)
    if ann_args is not None:
        ann_args = [arg for arg in ann_args if not isinstance(arg, typing.TypeVar)]
        if ret in ('Optional', 'Union') and len(ann_args) > 0 and ann_args[-1] is type(None):
            if len(ann_args) == 2:
                ann_args = ann_args[:-1]
            if len(ann_args) == 1:
                ret = 'Optional'
        if len(ann_args) > 0:
            args_anns = []
            for arg in ann_args:
                args_anns.append(make_annotation_string(arg))
            should_sort = ret == 'Union'
            ret = f'{ret}[{", ".join(sorted(args_anns) if should_sort else args_anns)}]'
    return ret



# -*- coding: future_annotations -*-
import itertools
from contextlib import contextmanager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Set, Tuple

    # avoid circular imports
    from nbsafety.tracing.trace_manager import BaseTraceManager


class TraceStack:
    def __init__(self, manager: BaseTraceManager):
        self._manager = manager
        self._stack: List[Tuple[Any, ...]] = []
        self._stack_item_initializers: Dict[str, Callable[[], Any]] = {}
        self._stack_items_with_manual_initialization: Set[str] = set()
        self._registering_stack_state_context = False
        self._field_mapping: Dict[str, int] = {}

    def _stack_item_names(self):
        return itertools.chain(self._stack_item_initializers.keys(), self._stack_items_with_manual_initialization)

    def get_field(self, field: str, depth: int = 1) -> Any:
        return self._stack[-depth][self._field_mapping[field]]

    @contextmanager
    def register_stack_state(self):
        self._registering_stack_state_context = True
        original_state = set(self._manager.__dict__.keys())
        yield
        self._registering_stack_state_context = False
        stack_item_names = set(self._manager.__dict__.keys() - original_state)
        for stack_item_name in stack_item_names - self._stack_items_with_manual_initialization:
            stack_item = self._manager.__dict__[stack_item_name]
            if isinstance(stack_item, TraceStack):
                self._stack_item_initializers[stack_item_name] = stack_item._clone
            elif stack_item is None:
                self._stack_item_initializers[stack_item_name] = lambda: None
            elif isinstance(stack_item, (int, bool, str, float)):
                init_val = type(stack_item)(stack_item)
                self._stack_item_initializers[stack_item_name] = lambda: init_val
            else:
                self._stack_item_initializers[stack_item_name] = type(stack_item)
        for i, stack_item_name in enumerate(self._stack_item_names()):
            self._field_mapping[stack_item_name] = i

    @contextmanager
    def needing_manual_initialization(self):
        assert self._registering_stack_state_context
        original_state = set(self._manager.__dict__.keys())
        yield
        self._stack_items_with_manual_initialization = set(self._manager.__dict__.keys() - original_state)

    @contextmanager
    def push(self):
        """
        Checks at the end of the context that everything requiring manual init was manually inited.
        """
        self._stack.append(tuple(self._manager.__dict__[stack_item] for stack_item in self._stack_item_names()))
        for stack_item, initializer in self._stack_item_initializers.items():
            self._manager.__dict__[stack_item] = initializer()
        for stack_item in self._stack_items_with_manual_initialization:
            del self._manager.__dict__[stack_item]
        yield
        uninitialized_items = []
        for stack_item in self._stack_items_with_manual_initialization:
            if stack_item not in self._manager.__dict__:
                uninitialized_items.append(stack_item)
        if len(uninitialized_items) > 0:
            raise ValueError(
                "Stack item(s) %s requiring manual initialization were not initialized" % uninitialized_items
            )

    def _clone(self):
        new_tracing_stack = TraceStack(self._manager)
        new_tracing_stack.__dict__ = dict(self.__dict__)
        new_tracing_stack._stack = []
        return new_tracing_stack

    def pop(self):
        for stack_item_name, stack_item in zip(self._stack_item_names(), self._stack.pop()):
            self._manager.__dict__[stack_item_name] = stack_item

    def clear(self):
        self._stack = []

    def __len__(self):
        return len(self._stack)





# -*- coding: future_annotations -*-
import ast
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Union

if TYPE_CHECKING:
    from nbsafety.data_model.scope import NamespaceScope


logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


# TODO: add generate return type to signature
def match_container_obj_or_namespace_with_literal_nodes(
    container_obj_or_namespace: Union[NamespaceScope, Dict[Any, Any], List[Any], Tuple[Any, ...]],
    literal_node: Union[ast.Dict, ast.List, ast.Tuple],
):
    try:
        gen = container_obj_or_namespace.items()  # type: ignore
        assert isinstance(literal_node, ast.Dict), 'got %s' % ast.dump(literal_node)
        yield from zip(gen, zip(literal_node.keys, literal_node.values))
    except (AttributeError, TypeError):
        assert isinstance(literal_node, (ast.List, ast.Tuple))
        elts = literal_node.elts
        cur_node = None
        cur_elt_idx = -1
        for i, obj_or_sym in enumerate(container_obj_or_namespace):
            if not isinstance(cur_node, ast.Starred) or len(elts) - cur_elt_idx - 1 >= len(container_obj_or_namespace) - i:
                cur_elt_idx += 1
                cur_node = elts[cur_elt_idx]
            yield (i, obj_or_sym), (None, cur_node)



# -*- coding: future_annotations -*-
import ast
import astunparse
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.ipython_utils import CellNotRunYetError
from nbsafety.singletons import nbs
from nbsafety.tracing.symbol_resolver import resolve_rval_symbols

if TYPE_CHECKING:
    from typing import Iterable, List, Optional, Set

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


USAGE = """Options:

[deps|show_deps|show_dependencies] <symbol_1>, <symbol_2> ...: 
    - This will print out the dependencies for given symbols.
      Multiple symbols should be separated with commas.

[stale|show_stale]: 
    - This will print out all the global variables that are stale. 

remove_dependency <parent_name> <child_name>:
    - This will remove the dependency between parent variable and the child variable.

slice <cell_num>:
    - This will print the code necessary to reconstruct <cell_num> using a dynamic
      program slicing algorithm.

add_dependency <parent_name> <child_name>:
    - This will add the dependency between parent variable and the child variable.

turn_off_warnings_for  <variable_name> <variable_name2> ...: 
    - This will turn off the warnings for given global variables. These variables will not be
      considered as stale anymore. Multiple variables should be separated with spaces.

turn_off_warnings_for  <variable_name> <variable_name2> ...: 
    - This will turn the warnings back on for given global variables. These variables could have
      stale dependencies now. Multiple variables should be separated with spaces."""


def show_deps(symbols: str) -> Optional[str]:
    usage = 'Usage: %safety show_[deps|dependencies] <symbol_1>[, <symbol_2> ...]'
    if len(symbols) == 0:
        logger.warning(usage)
        return None
    try:
        node = cast(ast.Expr, ast.parse(symbols).body[0]).value
    except SyntaxError:
        logger.warning('Could not find symbol metadata for %s', symbols)
        return None
    if isinstance(node, ast.Tuple):
        unresolved_symbols = node.elts
    else:
        unresolved_symbols = [node]
    statements = []
    for unresolved in unresolved_symbols:
        dsyms = resolve_rval_symbols(unresolved, should_update_usage_info=False)
        if len(dsyms) == 0:
            logger.warning('Could not find symbol metadata for %s', astunparse.unparse(unresolved).strip())
        for dsym in dsyms:
            parents = {par for par in dsym.parents if par.is_user_accessible}
            children: Set[DataSymbol] = set()
            children = children.union(
                *({child for child in children if child.is_user_accessible}
                  for children in dsym.children_by_cell_position.values())
            )
            if dsym.required_timestamp > 0:
                dsym_extra_info = 'last updated {}; required {}'.format(dsym.timestamp, dsym.required_timestamp)
            else:
                dsym_extra_info = 'defined in cell {}'.format(dsym.timestamp)
            statements.append(
                'Symbol {} ({}) is dependent on {} and is a parent of {}'.format(
                    dsym.full_namespace_path,
                    dsym_extra_info,
                    parents or 'nothing',
                    children or 'nothing',
                )
            )
    if len(statements) == 0:
        return None
    else:
        return '\n'.join(statements)


def show_stale(line_: str) -> Optional[str]:
    usage = 'Usage: %safety show_stale [global|all]'
    line = line_.split()
    if len(line) == 0 or line[0] == 'global':
        dsym_sets: Iterable[Iterable[DataSymbol]] = [nbs().global_scope.all_data_symbols_this_indentation()]
    elif line[0] == 'all':
        dsym_sets = nbs().aliases.values()
    else:
        logger.warning(usage)
        return None
    stale_set = set()
    for dsym_set in dsym_sets:
        for data_sym in dsym_set:
            if data_sym.is_stale and not data_sym.is_anonymous:
                stale_set.add(data_sym)
    if not stale_set:
        return 'No symbol has stale dependencies for now!'
    else:
        return 'Symbol(s) with stale dependencies are: %s' % stale_set


def trace_messages(line_: str) -> None:
    line = line_.split()
    usage = 'Usage: %safety trace_messages [enable|disable]'
    if len(line) != 1:
        logger.warning(usage)
        return
    setting = line[0].lower()
    if setting == 'on' or setting.startswith('enable'):
        nbs().trace_messages_enabled = True
    elif setting == 'off' or setting.startswith('disable'):
        nbs().trace_messages_enabled = False
    else:
        logger.warning(usage)


def set_highlights(cmd: str, rest: str) -> None:
    usage = 'Usage: %safety [hls|nohls]'
    if cmd == 'hls':
        nbs().mut_settings.highlights_enabled = True
    elif cmd == 'nohls':
        nbs().mut_settings.highlights_enabled = False
    else:
        rest = rest.lower()
        if rest == 'on' or rest.startswith('enable'):
            nbs().mut_settings.highlights_enabled = True
        elif rest == 'off' or rest.startswith('disable'):
            nbs().mut_settings.highlights_enabled = False
        else:
            logger.warning(usage)


def make_slice(line: str) -> Optional[str]:
    usage = 'Usage: %safety slice <cell_num>'
    try:
        cell_num = int(line)
    except:
        logger.warning(usage)
        return None
    try:
        deps = list(nbs().get_cell_dependencies(cell_num).items())
        deps.sort()
        return '\n\n'.join(f'# Cell {cell_num}\n' + content for cell_num, content in deps)
    except CellNotRunYetError:
        logger.warning("Cell %d has not yet been run", cell_num)
    return None


def _find_symbols(syms: List[str]) -> List[DataSymbol]:
    results = []
    for sym in syms:
        result = nbs().global_scope.lookup_data_symbol_by_name(sym)
        if result is None:
            logger.warning('Could not find symbol metadata for %s', sym)
        results.append(result)
    return results


def remove_dep(line_: str) -> None:
    usage = 'Usage: %safety remove_dependency <parent_name> <child_name>'
    line = line_.split()
    if len(line) != 2:
        logger.warning(usage)
        return
    results = _find_symbols(line)
    if len(results) != len(line):
        return
    parent_data_sym, child_data_sym = results
    if parent_data_sym not in child_data_sym.parents:
        logger.warning('The two symbols do not have a dependency relation')
        return
    for children in parent_data_sym.children_by_cell_position.values():
        children.remove(child_data_sym)
    child_data_sym.parents.remove(parent_data_sym)


def add_dep(line_: str) -> None:
    usage = 'Usage: %safety add_dependency <parent_name> <child_name>'
    line = line_.split()
    if len(line) != 2:
        logger.warning(usage)
        return
    results = _find_symbols(line)
    if len(results) != len(line):
        return
    parent_data_sym, child_data_sym = results
    if parent_data_sym in child_data_sym.parents:
        logger.warning('The two symbols already have a dependency relation')
        return
    parent_data_sym.children_by_cell_position[-1].add(child_data_sym)
    child_data_sym.parents.add(parent_data_sym)


def turn_off_warnings_for(line_: str) -> None:
    usage = 'Usage: %safety turn_off_warnings_for <variable_name> <variable_name2> ...'
    line = line_.split()
    if len(line) <= 1:
        logger.warning(usage)
        return
    for data_sym_name in line:
        data_sym = nbs().global_scope.lookup_data_symbol_by_name(data_sym_name)
        if data_sym:
            data_sym.disable_warnings = True
            logger.warning('Warnings are turned off for %s', data_sym_name)
        else:
            logger.warning('Could not find symbol metadata for %s', data_sym_name)


def turn_on_warnings_for(line_: str) -> None:
    usage = 'Usage: %safety turn_on_warnings_for <variable_name> <variable_name2> ...'
    line = line_.split()
    if len(line) == 0:
        logger.warning(usage)
        return
    for data_sym_name in line:
        data_sym = nbs().global_scope.lookup_data_symbol_by_name(data_sym_name)
        if data_sym:
            data_sym.disable_warnings = False
            logger.warning('Warnings are turned on for %s', data_sym_name)
        else:
            logger.warning('Could not find symbol metadata for %s', data_sym_name)



# -*- coding: future_annotations -*-
import ast
import itertools
import logging
from typing import cast, TYPE_CHECKING

from nbsafety.analysis.attr_symbols import get_attrsub_symbol_chain, AttrSubSymbolChain, CallPoint
from nbsafety.analysis.mixins import SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.scope import Scope

if TYPE_CHECKING:
    from typing import List, Optional, Set, Tuple, Union
    from nbsafety.types import SymbolRef

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


# TODO: have the logger warnings additionally raise exceptions for tests
class ComputeLiveSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitListsMixin, ast.NodeVisitor):
    def __init__(self, scope: Optional[Scope] = None, init_killed: Optional[Set[str]] = None):
        self.live: Set[SymbolRef] = set()
        self.scope = scope
        if init_killed is None:
            self.dead: Set[SymbolRef] = set()
        else:
            self.dead = cast('Set[SymbolRef]', init_killed)
        # TODO: use the ast context instead of hacking our own (e.g. ast.Load(), ast.Store(), etc.)
        self.in_kill_context = False
        self.inside_attrsub = False
        self.skip_simple_names = False

    def __call__(self, module_node: ast.Module):
        """
        This function should be called when we want to do a liveness check on a
        cell's corresponding ast.Module. For each line/block of the cell we
        first run the check of new assignments, then we obtain all the names.
        In these names, we put the ones that are user defined and not in the
        killed set into the return check_set for further checks.
        """
        # TODO: this will break if we ref a variable in a loop before killing it in the
        #   same loop, since we will add everything on the LHS of an assignment to the killed
        #   set before checking the loop body for live variables
        for node in module_node.body:
            self.visit(node)
        return self.live, self.dead

    def kill_context(self):
        return self.push_attributes(in_kill_context=True)

    def live_context(self):
        return self.push_attributes(in_kill_context=False)

    def attrsub_context(self, inside=True):
        return self.push_attributes(inside_attrsub=inside, skip_simple_names=inside)

    def args_context(self):
        return self.push_attributes(skip_simple_names=False)

    def _add_attrsub_to_live_if_eligible(self, ref: AttrSubSymbolChain):
        if ref in self.dead:
            return
        if len(ref.symbols) == 0:
            # can happen if user made syntax error like [1, 2, 3][4, 5, 6] (e.g. forgot comma)
            return
        leading_symbol = ref.symbols[0]
        if isinstance(leading_symbol, str) and leading_symbol in self.dead:
            return
        if isinstance(leading_symbol, CallPoint) and leading_symbol.symbol in self.dead:
            return
        self.live.add(ref)

    # the idea behind this one is that we don't treat a symbol as dead
    # if it is used on the RHS of an assignment
    def visit_Assign_impl(self, targets, value, aug_assign_target=None):
        this_assign_live = set()
        this_assign_dead = set()
        with self.push_attributes(live=this_assign_live):
            self.visit(value)
            if aug_assign_target is not None:
                self.visit(aug_assign_target)
        with self.push_attributes(dead=this_assign_dead):
            with self.kill_context():
                for target in targets:
                    self.visit_Assign_target(target)
        # TODO: ideally under the current abstraction we should
        #  not be resolving static references to symbols here
        if (
            self.scope is not None
            and len(this_assign_live) == 1
            and len(this_assign_dead) == 1
            and not (this_assign_dead <= self.dead)
        ):
            lhs, rhs = [
                get_symbols_for_references(x, self.scope, only_add_successful_resolutions=True)[0]
                for x in (this_assign_dead, this_assign_live)
            ]
            if len(lhs) == 1 and len(rhs) == 1:
                lhs, rhs = [next(iter(x)) for x in (lhs, rhs)]
                # hack to avoid marking `b` as live when objects are same,
                # or when it was detected that rhs symbol wasn't actually updated
                if lhs.obj is rhs.obj:
                    # it's a no-op, so treat it as such
                    this_assign_live.clear()
                    this_assign_dead.clear()
        this_assign_dead -= this_assign_live
        self.live |= this_assign_live
        self.dead |= this_assign_dead

    def visit_NamedExpr(self, node):
        self.visit_Assign_impl([node.target], node.value)

    def visit_Assign(self, node: ast.Assign):
        self.visit_Assign_impl(node.targets, node.value)

    def visit_AnnAssign(self, node: ast.AnnAssign):
        self.visit_Assign_impl([node.target], node.value)

    def visit_AugAssign(self, node: ast.AugAssign):
        self.visit_Assign_impl([], node.value, aug_assign_target=node.target)

    def visit_Assign_target(
        self, target_node: Union[ast.Attribute, ast.Name, ast.Subscript, ast.Tuple, ast.List, ast.expr]
    ):
        if isinstance(target_node, ast.Name):
            self.dead.add(target_node.id)
        elif isinstance(target_node, (ast.Attribute, ast.Subscript)):
            self.dead.add(get_attrsub_symbol_chain(target_node))
            if isinstance(target_node, ast.Subscript):
                with self.live_context():
                    self.visit(target_node.slice)
        elif isinstance(target_node, (ast.Tuple, ast.List)):
            for elt in target_node.elts:
                self.visit_Assign_target(elt)
        elif isinstance(target_node, ast.Starred):
            self.visit_Assign_target(target_node.value)
        else:
            logger.warning('unsupported type for node %s' % target_node)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.generic_visit(node.args.defaults)
        self.generic_visit(node.decorator_list)
        self.dead.add(node.name)

    def visit_Name(self, node):
        if self.in_kill_context:
            self.dead.add(node.id)
        elif not self.skip_simple_names and node.id not in self.dead:
            self.live.add(node.id)

    def visit_Tuple_or_List(self, node):
        for elt in node.elts:
            self.visit(elt)

    def visit_List(self, node):
        self.visit_Tuple_or_List(node)

    def visit_Tuple(self, node):
        self.visit_Tuple_or_List(node)

    def visit_For(self, node: ast.For):
        # Case "for a,b in something: "
        self.visit(node.iter)
        with self.kill_context():
            self.visit(node.target)

        for line in node.body:
            self.visit(line)

    def visit_ClassDef(self, node: ast.ClassDef):
        self.generic_visit(node.bases)
        self.generic_visit(node.decorator_list)

    def visit_Call(self, node: ast.Call):
        with self.args_context():
            self.generic_visit(node.args)
            for kwarg in node.keywords:
                self.visit(kwarg.value)
        self._add_attrsub_to_live_if_eligible(get_attrsub_symbol_chain(node))
        if isinstance(node.func, (ast.Attribute, ast.Subscript)):
            with self.attrsub_context():
                self.visit(node.func)
        else:
            self.visit(node.func)

    def visit_Attribute(self, node: ast.Attribute):
        if not self.inside_attrsub:
            self._add_attrsub_to_live_if_eligible(get_attrsub_symbol_chain(node))
        with self.attrsub_context():
            self.visit(node.value)

    def visit_Subscript(self, node: ast.Subscript):
        if not self.inside_attrsub:
            self._add_attrsub_to_live_if_eligible(get_attrsub_symbol_chain(node))
        with self.attrsub_context():
            self.visit(node.value)
        with self.attrsub_context(inside=False):
            self.visit(node.slice)

    def visit_GeneratorExp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_DictComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_ListComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_SetComp(self, node):
        self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)

    def visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(self, node):
        # TODO: as w/ for loop, this will have false positives on later live references
        for gen in node.generators:
            self.visit(gen.iter)
            with self.kill_context():
                self.visit(gen.target)
        # visit the elt at the end to ensure we don't add it to live vars if it was one of the generator targets
        self.visit(node.elt)

    def visit_Lambda(self, node):
        with self.kill_context():
            self.visit(node.args)

    def visit_arg(self, node):
        if self.in_kill_context:
            self.dead.add(node.arg)
        elif not self.skip_simple_names and node.arg not in self.dead:
            self.live.add(node.arg)


def get_symbols_for_references(
        symbol_refs: Set[SymbolRef],
        scope: Scope,
        only_add_successful_resolutions: bool = False,
) -> Tuple[Set[DataSymbol], Set[DataSymbol]]:
    symbols = set()
    called_symbols = set()
    for symbol_ref in symbol_refs:
        success = True
        if isinstance(symbol_ref, str):
            dsym = scope.lookup_data_symbol_by_name(symbol_ref)
            called_dsym = None
        elif isinstance(symbol_ref, AttrSubSymbolChain):
            dsym, called_dsym, success = scope.get_most_specific_data_symbol_for_attrsub_chain(symbol_ref)
        else:
            logger.warning('invalid type for ref %s', symbol_ref)
            continue
        if dsym is not None:
            if success or not only_add_successful_resolutions:
                symbols.add(dsym)
        if called_dsym is not None:
            called_symbols.add(called_dsym)
    return symbols, called_symbols


def compute_call_chain_live_symbols_and_cells(live: Set[DataSymbol]) -> Tuple[Set[DataSymbol], Set[int]]:
    seen = set()
    worklist = list(live)
    live = set(live)
    while len(worklist) > 0:
        called_dsym = worklist.pop()
        if called_dsym in seen:
            continue
        # TODO: handle callable classes
        if not called_dsym.is_function:
            continue
        seen.add(called_dsym)
        live_refs, _ = compute_live_dead_symbol_refs(
            cast(ast.FunctionDef, called_dsym.stmt_node).body, init_killed=set(called_dsym.get_definition_args())
        )
        live_symbols, called_symbols = get_symbols_for_references(live_refs, called_dsym.call_scope)
        worklist.extend(called_symbols)
        live |= {
            sym for sym in itertools.chain(live_symbols, called_symbols) if sym.is_globally_accessible
        }
    return live, {called_dsym.timestamp for called_dsym in seen}


def compute_live_dead_symbol_refs(
    code: Union[ast.Module, List[ast.stmt], str],
    scope: Scope = None,
    init_killed: Optional[Set[str]] = None,
) -> Tuple[Set[SymbolRef], Set[SymbolRef]]:
    if init_killed is None:
        init_killed = set()
    if isinstance(code, str):
        code = ast.parse(code)
    elif isinstance(code, list):
        code = ast.Module(body=code)
    return ComputeLiveSymbolRefs(scope=scope, init_killed=init_killed)(code)
